package djmil.cordacheckers.cordaclient;

import static java.util.Objects.requireNonNull;

import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.ObjectMapper;

import djmil.cordacheckers.cordaclient.dao.GameView;
import djmil.cordacheckers.cordaclient.dao.HoldingIdentity;
import djmil.cordacheckers.cordaclient.dao.Stone;
import djmil.cordacheckers.cordaclient.dao.Rank;
import djmil.cordacheckers.cordaclient.dao.VirtualNode;
import djmil.cordacheckers.cordaclient.dao.VirtualNodeList;
import djmil.cordacheckers.cordaclient.dao.flow.RequestBody;
import djmil.cordacheckers.cordaclient.dao.flow.ResponseBody;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.Req;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.Rsp;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.ReqGameBoardMove;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.ReqGameProposalCreate;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.RspGameState;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.RspGameStateList;
import djmil.cordacheckers.cordaclient.dao.flow.arguments.RspRankMap;

@Service
public class CordaClient {

    @Value("${corda.client.maxPollAttempts}")
    private int maxPollAttempts;

    private final RestTemplate restTemplate;
    private final ObjectMapper jsonMapper;

    public CordaClient(RestTemplateBuilder restTemplateBuilder , ObjectMapper jsonMapper) {
        this.restTemplate = restTemplateBuilder.build();
        this.jsonMapper = jsonMapper;
    }

    public List<VirtualNode> fetchVirtualNodeList() {
        ResponseEntity<VirtualNodeList> resp = this.restTemplate.exchange(
            "/virtualnode",
            HttpMethod.GET, 
            null, 
            VirtualNodeList.class
        );

        if (resp.getStatusCode() != HttpStatus.OK) {
            throw new RuntimeException("CordaClient.getVirtualNodeList: unexpected request status "+resp.getStatusCode()) ;
        }

        return requireNonNull(
                resp.getBody(),
                "CordaClient.getVirtualNodeList: empty getBody()"
            ).virtualNodes();
    }

    // use custodian holding identity to get ranking table of all players
    public Map<String, Rank> fetchRanking(HoldingIdentity holdingIdentity) {
        final RequestBody requestBody = new RequestBody(
            "ranking-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameresult.RankingFlow",
            new Req());

        return cordaFlowExecute(holdingIdentity, requestBody, RspRankMap.class)
            .getResponce(requestBody);
    }

    /**
     * @param holdingIdentity
     * @return list of unconsumed (active) GameStateViews
     */
    public List<GameView> gameStateList(HoldingIdentity holdingIdentity) {
        final RequestBody requestBody = new RequestBody(
            "gs.list-" + UUID.randomUUID(),
            "djmil.cordacheckers.gamestate.ListFlow",
            new Req());

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameStateList.class)
            .getResponce(requestBody);
    }

    public GameView gameStateGet(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gs.get-" + UUID.randomUUID(),
            "djmil.cordacheckers.gamestate.GetFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameProposalCreate(HoldingIdentity issuer, HoldingIdentity acquier, Stone.Color acquierColor,
        String message) {
        
        return gameProposalCreate(issuer, acquier, acquierColor, GameView.defaultGameBoard, message);
    }

    public GameView gameProposalCreate(HoldingIdentity issuer, HoldingIdentity acquier, Stone.Color acquierColor,
            Map<Integer, Stone> board, String message) {
        final RequestBody requestBody = new RequestBody(
            "gp.create-" + UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.CreateFlow",
            new ReqGameProposalCreate(
                acquier.x500Name(),
                acquierColor,
                board,
                message));

        return cordaFlowExecute(issuer, requestBody,RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameProposalReject(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gp.reject-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.RejectFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameProposalCancel(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gp.cancel-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.CancelFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameProposalAccept(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gp.accept-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameproposal.AcceptFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameBoardSurrender(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gb.surrender-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.SurrenderFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameBoardMove(HoldingIdentity holdingIdentity, UUID gameUuid, List<Integer> move,
            String message) {
        final RequestBody requestBody = new RequestBody(
            "gb.move-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.MoveFlow",
            new ReqGameBoardMove(
                move, gameUuid, message));

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameDrawRequest(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gd.request-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.DrawRequestFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameDrawAccept(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gd.accept-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.DrawAcceptFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    public GameView gameDrawReject(HoldingIdentity holdingIdentity, UUID gameUuid) {
        final RequestBody requestBody = new RequestBody(
            "gd.reject-" +UUID.randomUUID(),
            "djmil.cordacheckers.gameboard.DrawRejectFlow",
            gameUuid);

        return cordaFlowExecute(holdingIdentity, requestBody, RspGameState.class)
            .getResponce(requestBody);
    }

    private <T extends Rsp<?>> T cordaFlowExecute(HoldingIdentity holdingIdentity, RequestBody requestBody, Class<T> flowReponceType) {
        try {
            final String requestBodyJson = this.jsonMapper
                .writeValueAsString(requestBody);

            final ResponseBody startedFlow = cordaFlowPost(holdingIdentity, requestBodyJson);

            final String flowResponce = cordaFlowPoll(startedFlow);

            return this.jsonMapper
                .readValue(flowResponce, flowReponceType);
        } 
        catch (Exception e) {
            throw new RuntimeException("Unable to perform "+requestBody.flowClassName()
                                        +". Reason "+e.getMessage());
        }
    }

    private ResponseBody cordaFlowPost(HoldingIdentity holdingIdentity, String requestBodyJson) {
		final HttpEntity<String> request = new HttpEntity<>(requestBodyJson);

        final ResponseEntity<ResponseBody> responce = this.restTemplate.exchange(
            "/flow/" + holdingIdentity.shortHash(),
            HttpMethod.POST,
            request, 
            ResponseBody.class
        );
        
        if (!responce.getStatusCode().is2xxSuccessful()) {
            throw new RuntimeException("CordaClient.startCordaFlow: unexpected request status "
                                        +responce.getStatusCode()) ;
        }

        final ResponseBody responseBody = requireNonNull(
            responce.getBody(),
            "CordaClient.startCordaFlow: empty getBody()"
        );

        if (!responseBody.isFlowStarted()) {
            throw new RuntimeException("CordaClient.startCordaFlow: Unexpected status: "
                +responseBody.flowStatus() + ": " 
                +responseBody.flowError());
        }

        return responseBody;
    }

    private String cordaFlowPoll(ResponseBody startedFlow) throws InterruptedException {

        for (int retry = 0; retry < maxPollAttempts; retry++) {
            // Give Corda cluster some time to process our request
            final int waitSec = retry*retry +2 < 5 ? retry*retry +2 : 5; // a progression of 2 3 6 10 18 27 ..
            TimeUnit.SECONDS.sleep(waitSec);

            final ResponseEntity<ResponseBody> responce = this.restTemplate.exchange(
                "/flow/" 
                    + startedFlow.holdingIdentityShortHash()+"/"
                    + startedFlow.clientRequestId(),
                HttpMethod.GET,
                null, 
                ResponseBody.class
            );
          
            if (responce.getStatusCode() != HttpStatus.OK) {
                throw new RuntimeException("CordaClient.cordaFlowPoll: unexpected request status "
                                            +responce.getStatusCode()) ;
            }

            final ResponseBody responseBody = requireNonNull(
                responce.getBody(),
                "CordaClient.cordaFlowPoll: empty getBody()"
            );

            if (responseBody.flowError() != null) {
                throw new RuntimeException("CordaClient.cordaFlowPoll: flow execution error: " 
                                            +responseBody.flowError());
            }

            if (responseBody.isFlowCompleted() && responseBody.flowResult() != null) {
                return responseBody.flowResult();
            }
        }

        throw new RuntimeException ("CordaClient.cordaFlowPoll: retry limit");
    }

}
