import './Checkers.css'
import React, { useState } from 'react'

export const Color = {
    white: "WHITE",
    black: "BLACK",

    opposite: (color) => {
        if (color === Color.white)
            return Color.black;
        if (color === Color.black)
            return Color.white;

        return color;
    }
};

/*
 * Stone
 */
export function Stone({ color, type, move }) {
    const style = !move ? null : {
        position: 'absolute',
        left: move[0],
        top: move[1],
    }

    switch (color) {
        case Color.white:
            return <WhiteStone type={type} style={style} />;

        case Color.black:
            return <BlackStone type={type} style={style} />;

        case '':
        case undefined:
        case null:
            return; // no stone :)

        default:
            console.warn('Unknown color', color)
    }
}

export function WhiteStone({ type, style }) {
    if (type === 'KING')
        return <span className='Stone white king' style={style} >⛁</span>
    else
        return <span className='Stone white man' style={style}>⛀</span>
}

export function BlackStone({ type, style }) {
    if (type === 'KING')
        return <span className='Stone black king' style={style}>⛃</span>
    else
        return <span className='Stone black man' style={style}>⛂</span>
}

/*
 * Player
 */
export function Player({ color, name }) {
    return (
        <div className='Player'>
            {name}
            <Stone color={color} />
        </div>
    )
}

/*
 * Board
 */
export function Board({ board, flip, onStoneClick, onStoneMove }) {
    const [[moveId, moveX, moveY], setMove] = useState([0, 0, 0]);

    if (!board)
        board = [];

    const isInteractive = (typeof onStoneClick === 'function' || typeof onStoneMove === 'function') ? ' interactive' : '';

    const WhiteTile = ({ id }) => {
        const stone = board[id];

        return (
            <div className={'Tile white' + isInteractive + (id === moveId ? ' selected' : '')}
                onClick={!onStoneClick ? null :
                    () => onStoneClick(id)
                }

                onMouseDown={!onStoneMove || !stone ? null :
                    (e) => setMove([id, e.clientX, e.clientY])
                }

                onMouseUp={!onStoneMove || !moveId ? null :
                    () => { onStoneMove([moveId, id]); setMove([0, 0, 0]) }
                }
            >
                <Stone color={stone?.color} type={stone?.type} />
            </div>
        );
    }

    const BlackTile = () => {
        return <div className='Tile black' />
    }

    const movingStone = board[moveId];

    return (
        <div className={'Board' + (flip ? ' flip' : '')}
            onMouseMove={!moveId ? null :
                (e) => e.buttons ? setMove([moveId, e.clientX, e.clientY]) : setMove([0, 0, 0])
            }
        >

            {!movingStone ? null :
                <Stone color={movingStone.color} type={movingStone.type} move={[moveX, moveY]} />
            }

            <div className='row'>
                <BlackTile /> <WhiteTile id={1} />
                <BlackTile /> <WhiteTile id={2} />
                <BlackTile /> <WhiteTile id={3} />
                <BlackTile /> <WhiteTile id={4} />
            </div>
            <div className='row'>
                <WhiteTile id={5} /> <BlackTile />
                <WhiteTile id={6} /> <BlackTile />
                <WhiteTile id={7} /> <BlackTile />
                <WhiteTile id={8} /> <BlackTile />
            </div>
            <div className='row'>
                <BlackTile /> <WhiteTile id={9} />
                <BlackTile /> <WhiteTile id={10} />
                <BlackTile /> <WhiteTile id={11} />
                <BlackTile /> <WhiteTile id={12} />
            </div>
            <div className='row'>
                <WhiteTile id={13} /> <BlackTile />
                <WhiteTile id={14} /> <BlackTile />
                <WhiteTile id={15} /> <BlackTile />
                <WhiteTile id={16} /> <BlackTile />
            </div>

            <div className='row'>
                <BlackTile /> <WhiteTile id={17} />
                <BlackTile /> <WhiteTile id={18} />
                <BlackTile /> <WhiteTile id={19} />
                <BlackTile /> <WhiteTile id={20} />
            </div>
            <div className='row'>
                <WhiteTile id={21} /> <BlackTile />
                <WhiteTile id={22} /> <BlackTile />
                <WhiteTile id={23} /> <BlackTile />
                <WhiteTile id={24} /> <BlackTile />
            </div>
            <div className='row'>
                <BlackTile /> <WhiteTile id={25} />
                <BlackTile /> <WhiteTile id={26} />
                <BlackTile /> <WhiteTile id={27} />
                <BlackTile /> <WhiteTile id={28} />
            </div>
            <div className='row'>
                <WhiteTile id={29} /> <BlackTile />
                <WhiteTile id={30} /> <BlackTile />
                <WhiteTile id={31} /> <BlackTile />
                <WhiteTile id={32} /> <BlackTile />
            </div>
        </div>
    )
}

const blackMan = { color: Color.black, type: 'MAN' };
const blackKing = { color: Color.black, type: 'KING' };
const whiteMan = { color: Color.white, type: 'MAN' };
const whiteKing = { color: Color.white, type: 'KING' };

export const defaultBoard = {
    1: blackMan, 2: blackMan, 3: blackMan, 4: blackMan, 5: blackMan, 6: blackMan, 7: blackMan, 8: blackMan, 9: blackMan, 10: blackMan, 11: blackMan, 12: blackMan,
    21: whiteMan, 22: whiteMan, 23: whiteMan, 24: whiteMan, 25: whiteMan, 26: whiteMan, 27: whiteMan, 28: whiteMan, 29: whiteMan, 30: whiteMan, 31: whiteMan, 32: whiteMan,
}

export function nextStone(currentStone) {
    const color = currentStone?.color;
    const type = currentStone?.type;

    if (!color || !type)
        return blackMan;

    if (color === Color.black && type === 'MAN')
        return blackKing;

    if (color === Color.black && type === 'KING')
        return whiteMan;

    if (color === Color.white && type === 'MAN')
        return whiteKing;

    if (color === Color.white && type === 'KING')
        return null;
}