import React from 'react';
import { GamesContext } from '../context/games';
import { NavLink, Routes, Route } from 'react-router-dom';

import NewGame from './games/view/NewGame';
import { GameProposalSelector, ActiveGameSelector, GameArchiveSelector } from './games/view/GameSelector';
import { Create, Accept, Reject, Cancel, DrawRequest, DrawAccept, DrawReject, Surrender, Backward, Forward } from './games/ActionPanel';
import GameBoard from './games/GameBoard';
import { nextStone } from '../components/Checkers';
import Message2Opponent from './games/Message2Opponent';
import Counter from '../components/Counter';

import './Games.css';

export default function Games({ context: { gamesReducer, gamesApi }, players }) {
  const [games, dispatchGames] = gamesReducer;

  return (
    <GamesContext.Provider value={games} >
      <div className='Games'>

        <div className='left-side'>
          <ViewSelector games={games} />
          <ViewProvider dispatchGames={dispatchGames} players={players} />
        </div>

        <div className='right-side'>
          <ActionPanel gamesApi={gamesApi} />
          {/* <GameMessage /> */}
          <GameBoardRoutes gamesReducer={gamesReducer} gamesApi={gamesApi} username={players.currentUser} />
          <Message2Opponent dispatchGames={dispatchGames} />
        </div>

      </div >
    </GamesContext.Provider>
  )
};

function ViewSelector({ games }) {
  const awaiting = countGames(games.gamesList);

  return (
    <nav className='ViewSelector' >
      <NavLink to='new'>New</NavLink>
      <NavLink to='proposal'>Proposal<Counter number={awaiting.proposals} /></NavLink>
      <NavLink to='active'  >Active<Counter number={awaiting.active} /></NavLink>
      <NavLink to='archive' >Archive</NavLink>
    </nav>
  )
}

function ViewProvider({ dispatchGames, players }) {

  return (
    <div className='ViewProvider'>
      <Routes>

        <Route path='new' element={
          <NewGame setPlayers={(opponentName, myColor) => dispatchGames({ type: 'nextNewGame', opponentName, myColor })}
            players={players}
          />
        } />

        <Route path='proposal' element={
          <GameProposalSelector onSelect={(selectedUUID) => dispatchGames({ type: 'nextProposal', selectedUUID })} />
        } />

        <Route path='active' element={
          <ActiveGameSelector onSelect={(selectedUUID) => dispatchGames({ type: 'nextActive', selectedUUID })} />
        } />

        <Route path='archive' element={
          <GameArchiveSelector onSelect={(selectedUUID) => dispatchGames({ type: 'nextArchive', selectedUUID })} />
        } />

      </Routes>
    </div>
  )
}

function ActionPanel({ gamesApi }) {
  return (
    <div className='ActionPanel'>
      <Routes>

        <Route path='new' element={
          <Create onClick={(reqParams) => gamesApi.pushNewGame(reqParams)} />
        } />

        <Route path='proposal' element={[
          <Accept key={1} onClick={(uuid) => gamesApi.pushGameProposalAccept({ uuid })} />,
          <Reject key={2} onClick={(uuid) => gamesApi.pushGameProposalReject({ uuid })} />,
          <Cancel key={3} onClick={(uuid) => gamesApi.pushGameProposalCancel({ uuid })} />
        ]} />

        <Route path='active' element={[
          <DrawRequest key={1} onClick={(uuid) => gamesApi.pushGameDrawRequest({ uuid })} />,
          <DrawAccept key={2} onClick={(uuid) => gamesApi.pushGameDrawAccept({ uuid })} />,
          <DrawReject key={3} onClick={(uuid) => gamesApi.pushGameDrawReject({ uuid })} />,
          <Surrender key={4} onClick={(uuid) => gamesApi.pushGameSurrender({ uuid })} />
        ]} />

        <Route path='archive' element={[
          <Backward key={1} />,
          <Forward key={2} />
        ]} />

      </Routes>
    </div>
  )
}

function GameBoardRoutes({ gamesReducer, gamesApi, username }) {
  const [games, dispatchGames] = gamesReducer;

  const onStoneClick = (uuid, cellId) => {
    let board = { ...games.newGame.board };
    board[cellId] = nextStone(board[cellId]);
    dispatchGames({ type: 'nextNewGame', board });
  }

  const onStoneMove = (uuid, move) => gamesApi.pushGameMove({uuid, move, message: games.active.message});

  return (
    <Routes>
      <Route path='new' element={<GameBoard username={username} onStoneClick={onStoneClick} />} />
      <Route path='proposal' element={<GameBoard username={username} />} />
      <Route path='active' element={<GameBoard username={username} onStoneMove={onStoneMove} />} />
      <Route path='archive' element={<GameBoard username={username} />} />
    </Routes>
  )
}

function countGames(gamesList) {

  var awaiting = {
    proposals: 0,
    active: 0
  };

  if (!gamesList)
    return awaiting;

  for (const game of gamesList) {
    switch (game.status) {
      case 'GAME_PROPOSAL_WAIT_FOR_YOU':
        awaiting.proposals++;
        break;
      case 'GAME_BOARD_WAIT_FOR_YOU':
      case 'DRAW_REQUEST_WAIT_FOR_YOU':
        awaiting.active++;
        break;
      default:
        break;
    }
  }

  return awaiting;
}